import re
from collections import namedtuple
from urllib.parse import urlencode

import requests


class RtTicket(namedtuple('RtTicket', ('number', 'owner', 'subject', 'queue', 'status'))):
    """A namedtuple representing an RT ticket."""

    def __str__(self):
        return (
            'rt#{self.number}: "{self.subject}" | '
            '{self.queue}, {self.status}, owned by {self.owner} | '
            'https://ocf.io/rt/{self.number}'
        ).format(self=self)

    @classmethod
    def from_number(cls, connection, num):
        resp = connection.get('https://rt.ocf.berkeley.edu/REST/1.0/ticket/{}/view'.format(num))
        assert resp.status_code == 200, resp.status_code
        assert '200 Ok' in resp.text

        lines = resp.text.splitlines()

        def find(header):
            for line in lines:
                if line.startswith(header + ': '):
                    return line.split(': ', 1)[1]

        return cls(
            number=num,
            owner=find('Owner'),
            subject=find('Subject'),
            queue=find('Queue'),
            status=find('Status'),
        )

    @classmethod
    def create(cls, connection, queue, requestor, subject, text, **kwargs):
        """Create an RT ticket and returns an instance of the result"""
        # RT prefixes multiline strings by a blank space
        text = text.replace('\n', '\n ')

        data = {
            'id': 'ticket/new',
            'Queue': queue,
            'Requestor': requestor,
            'Subject': subject,
            'Text': text,
            **kwargs,
        }

        # RT's incoming data format has the form key: value, but these aren't HTTP Headers
        body = ''
        for k, v in data.items():
            body += '{}: {}\n'.format(k, v)

        # RT requires the POST content to be sent with no filename hence 'content': (None, body)
        resp = connection.post('https://rt.ocf.berkeley.edu/REST/1.0/ticket/new', files={'content': (None, body)})
        resp.raise_for_status()
        assert '200 Ok' in resp.text

        match = re.search(r'Ticket ([0-9]+) created.', resp.text)
        assert match, '200 response but no ticket number found in RT response'

        ticket_number = int(match.group(1))
        return ticket_number


def rt_connection(user, password):
    """Return a requests Session object authenticated against RT.

    Currently, we only enable access to the REST API from a few select IPs.
    """
    s = requests.Session()
    resp = s.post(
        'https://rt.ocf.berkeley.edu/REST/1.0/',
        data=urlencode({'user': user, 'pass': password}),
        timeout=20,
    )
    assert resp.status_code == 200, resp.status_code
    assert '200 Ok' in resp.text
    return s


class RtCredentials(namedtuple('RtCredentials', [
    'username',
    'password',
])):
    """Credentials for programmatically accessing RT.

    :param username: str
    :param password: str
    """



import functools
import re

from ocflib.infra import mysql


SHORTURL_SLUG_ALLOWED_CHARS = r'^[\w./+:-]+$'
SHORTURL_REGEX = re.compile(SHORTURL_SLUG_ALLOWED_CHARS)

get_connection = functools.partial(
    mysql.get_connection,
    db='ocfshorturls',
    user='anonymous',
    password=None,
)


def _validate_slug(slug):
    if len(slug) > 100:
        raise ValueError('shorturl len is {}, must be less than 100'.format(len(slug)))
    if not SHORTURL_REGEX.search(slug):
        raise ValueError("shorturl '{}' contains illegal characters".format(slug))


def get_shorturl(ctx, slug):
    """Get the target of a shorturl by its slug."""

    query = 'SELECT `target` FROM `shorturls_public` WHERE `slug` = %s'
    ctx.execute(query, (slug))

    candidate = ctx.fetchone()
    return candidate['target'] if candidate else None


def add_shorturl(ctx, slug, target):
    """Add a shorturl to the database."""

    _validate_slug(slug)

    # we don't need to explicitly check for duplicates because
    # there's a uniqueness constraint on the slug column
    query = 'INSERT INTO `shorturls` (slug, target) VALUES (%s, %s)'
    ctx.execute(query, (slug, target))


def delete_shorturl(ctx, slug):
    """Delete a shorturl from the database."""

    query = 'DELETE FROM `shorturls` WHERE `slug` = %s'
    ctx.execute(query, (slug,))


def rename_shorturl(ctx, old_slug, new_slug):
    """Rename a shorturl."""

    _validate_slug(new_slug)

    query = 'UPDATE `shorturls` SET `slug` = %s WHERE `slug` = %s'
    ctx.execute(query, (new_slug, old_slug))


def replace_shorturl(ctx, slug, new_target):
    """Change the target of a shorturl."""

    query = 'UPDATE `shorturls` SET `target` = %s WHERE `slug` = %s'
    ctx.execute(query, (new_target, slug))



"""Random account methods that don't fit anywhere else."""
import grp
import os.path
import re

import pexpect

import ocflib.account.validators as validators
from ocflib.infra.ldap import OCF_LDAP_PEOPLE


def password_matches(username, password):
    """Returns True if the password matches the user account given"""

    validators.validate_username(username)
    validators.validate_password(username, password, strength_check=False)

    if not validators.user_exists(username):
        raise ValueError("User doesn't exist")

    cmd = 'kinit --no-forwardable -l0 {}@OCF.BERKELEY.EDU'.format(username)
    child = pexpect.spawn(cmd, timeout=10)

    child.expect("{}@OCF.BERKELEY.EDU's Password:".format(username))
    child.sendline(password)

    child.expect(pexpect.EOF)
    child.close()

    return child.exitstatus == 0


def extract_username_from_principal(principal):
    """Extract username from principal.

    >>> extract_username("ckuehl@OCF.BERKELEY.EDU")
    'ckuehl'

    >>> extract_username("ckuehl/admin@OCF.BERKELEY.EDU")
    'ckuehl'
    """

    REGEX = '^([a-z]{3,16})(/[a-z]*)?@OCF\\.BERKELEY\\.EDU$'
    match = re.match(REGEX, principal)

    if not match:
        raise ValueError('Invalid username')

    return match.group(1)


def home_dir(user):
    """Returns the user's home directory path."""
    return '/' + os.path.join('home', user[0], user[:2], user)


def web_dir(user):
    """Returns the user's web directory path."""
    return '/' + os.path.join('services', 'http', 'users', user[0], user)


def public_html_path(user):
    """Returns the user's public_html path."""
    return os.path.join(home_dir(user), 'public_html')


def is_in_group(user, group):
    """Return whether the user is in a group.

    Only will return True if the user is a supplementary member, so "sorry"
    won't work here.

    :param group: UNIX group to use.
    """
    return user in list_group(group)


def list_group(group):
    """Return a list of OCF users in a group

    Only returns users who have this group listed as a supplementary group, so
    "sorry" won't work here.

    :param group: UNIX group to list.
    """
    return grp.getgrnam(group).gr_mem


def dn_for_username(username):
    return 'uid={user},{base_people}'.format(
        user=username,
        base_people=OCF_LDAP_PEOPLE,
    )



import re
import uuid
from collections import namedtuple
from datetime import date
from textwrap import dedent

from ocflib.account.search import user_attrs
from ocflib.account.search import user_attrs_ucb
from ocflib.infra.github import GitRepo
from ocflib.infra.rt import rt_connection
from ocflib.infra.rt import RtTicket
from ocflib.misc.mail import send_problem_report

VHOST_DB_PATH = '/etc/ocf/vhost.conf'
GITHUB_VHOST_REPO = 'ocf/etc'
GITHUB_VHOST_WEB_PATH = 'configs/vhost.conf'


def get_vhost_db(remote=False):
    """Returns lines from the vhost config file. If remote is True, fetches it
    from GitHub."""
    if remote:
        vhosts = GitRepo(GITHUB_VHOST_REPO).get_file(GITHUB_VHOST_WEB_PATH)
        return vhosts.splitlines()
    else:
        with open(VHOST_DB_PATH) as f:
            return f.read().splitlines()


def pr_new_vhost(credentials, username, aliases=None, docroot=None, flags='', rt_ticket=''):
    """
    Creates a GitHub pull request on the vhosts file to add a new vhost
    """
    if not aliases:
        aliases = '-'
    else:
        aliases = ','.join(aliases)

    if not docroot:
        docroot = '-'

    repo = GitRepo(GITHUB_VHOST_REPO, credentials=credentials)
    vhosts_lines = get_vhost_db(remote=True)

    # skip the initial comment block
    idx = 0
    while idx < len(vhosts_lines):
        if not vhosts_lines[idx].startswith('#'):
            break
        idx += 1

    vhosts_lines.insert(idx, dedent("""
        # added {date} web {rt_ticket}
        {username} {aliases} {docroot} {flags}""").format(
        date=date.today(),
        rt_ticket='rt#{}'.format(rt_ticket) if rt_ticket else '',
        username=username,
        aliases=aliases,
        docroot=docroot,
        flags=flags,
    ))

    new_vhosts_file = '\n'.join(vhosts_lines) + '\n'  # newline at eof
    new_branch_name = '{rt_ticket}{id}'.format(
        rt_ticket='rt#{}-'.format(rt_ticket) if rt_ticket else '',
        id=uuid.uuid4().hex
    )

    repo.modify_and_branch(
        'master',
        new_branch_name,
        'rt#{rt_ticket}: Add vhost for {username}'.format(
            rt_ticket=rt_ticket,
            username=username,
        ),
        GITHUB_VHOST_WEB_PATH,
        new_vhosts_file,
    )

    pull_body = dedent("""
        Submitted from ocflib on {date}

        Username: {username}
        Aliases: {aliases}
        Document root: {docroot}
        Flags: {flags}

        Associated RT Ticket: rt#{rt_ticket}
        https://ocf.io/rt/{rt_ticket}
        """).format(
        date=date.today(),
        rt_ticket=rt_ticket,
        username=username,
        aliases=aliases,
        docroot=docroot,
        flags=flags,
    )

    repo.github.create_pull(
        title='rt#{rt_ticket}: Add vhost for {username}'.format(
            rt_ticket=rt_ticket,
            username=username,
        ),
        body=pull_body,
        base='master',
        head=new_branch_name,
    )


def get_vhosts():
    """Returns a list of virtual hosts in convenient format.

    >>> get_vhosts()
    ...
    {
        'bpreview.berkeley.edu': {
            'username': 'bpr',
            'aliases': ['bpr.berkeley.edu'],
            'docroot': '/',
            'flags': [],
        }
    }
    ...
    """
    def fully_qualify(host):
        """Fully qualifies a hostname (by appending .berkeley.edu) if it's not
        already fully-qualified."""
        return host if '.' in host else host + '.berkeley.edu'

    vhosts = {}

    for line in get_vhost_db():
        if not line or line.startswith('#'):
            continue

        fields = line.split(' ')

        if len(fields) < 5:
            flags = []
        else:
            flags = re.match(r'\[(.*)\]$', fields[4]).group(1).split(',')

        username, host, aliases, docroot = fields[:4]

        if aliases != '-':
            aliases = list(map(fully_qualify, aliases.split(',')))
        else:
            aliases = []

        vhosts[fully_qualify(username if host == '-' else host)] = {
            'username': username,
            'aliases': aliases,
            'docroot': '/' if docroot == '-' else docroot,
            'flags': flags,
        }

    return vhosts


def has_vhost(user):
    """Returns whether or not a virtual host is already configured for
    the given user."""
    return any(vhost['username'] == user for vhost in get_vhosts().values())


def eligible_for_vhost(user):
    """Returns whether a user account is eligible for virtual hosting.

    Currently, group accounts, faculty, and staff are eligible for virtual
    hosting.
    """
    attrs = user_attrs(user)
    if 'callinkOid' in attrs:
        return True
    elif 'calnetUid' in attrs:
        attrs_ucb = user_attrs_ucb(attrs['calnetUid'])
        # TODO: Uncomment when we get a privileged LDAP bind.
        if attrs_ucb:  # and 'EMPLOYEE-TYPE-ACADEMIC' in attrs_ucb['berkeleyEduAffiliations']:
            return True

    return False


def get_tasks(celery_app, credentials=None):
    """Return Celery tasks instantiated against the provided instance."""

    @celery_app.task
    def create_new_vhost(request):
        try:
            conn = rt_connection(credentials['rt'].username, credentials['rt'].password)
            ticket_number = RtTicket.create(
                conn,
                'hostmaster',
                request.requestor,
                request.subject,
                request.message,
            )

            pr_new_vhost(
                credentials['github'],
                request.username,
                request.aliases,
                request.docroot,
                request.flags,
                ticket_number,
            )
            return True
        except Exception as e:
            send_problem_report(str(e))
            return False

    return _VirtualHostTasks(
        create_new_vhost=create_new_vhost,
    )


_VirtualHostTasks = namedtuple('VirtualHostTasks', [
    'create_new_vhost',
])


class NewVirtualHostRequest(namedtuple('NewVirtualHostRequest', [
    'username',
    'requestor',
    'subdomain',
    'aliases',
    'docroot',
    'flags',
    'rt_ticket',
    'subject',
    'message',
])):
    """Request for account creation.

    :param username: str
    :param requestor: str
    :param subdomain: str
    :param aliases: str or None
    :param docroot: str or None
    :param flags: str or None
    :param rt_ticket: str or None
    :param subject: str or None
    :param message: str or None
    """
